Ext.namespace('dSS', 'dSS.data');

dSS.data.ZoneStore = Ext.extend(Ext.data.Store, {
	constructor: function(config) {
		var zoneRecord = Ext.data.Record.create([
			{name:"name"},
			{name:"id"},
			{name:"primary", defaultValue: false}
		]);

		var zoneReader = new Ext.data.JsonReader(
			{
				root: "zones"
			},
			zoneRecord
		);

		Ext.apply(this, { reader: zoneReader });
		dSS.data.ZoneStore.superclass.constructor.call(this, arguments);
	}
});
Ext.namespace('dSS', 'dSS.data');

dSS.data.ModulatorStore = Ext.extend(Ext.data.Store, {
	constructor: function(config) {
		var modulatorRecord = Ext.data.Record.create([
			{name:"busId", mapping: "busid"},
			{name:"deviceType"},
			{name:"dsid"},
			{name:"hwName"},
			{name:"hwVersion"},
			{name:"isPresent"},
			{name:"name"},
			{name:"swVersion"},
            {name:"consumption"},
            {name:"measureDate"}
		]);

		var modulatorReader = new Ext.data.JsonReader(
			{
				root: "result.circuits"
			},
			modulatorRecord
		);

		Ext.apply(this, { reader: modulatorReader, url: '/json/apartment/getCircuits' });
		dSS.data.ModulatorStore.superclass.constructor.call(this, arguments);

        this.on(
            'load',
            function(records, options) {
                var store = this;
                Ext.Ajax.request({
                    url: '/json/metering/getLatest',
                    disableCaching: true,
                    method: 'GET',
                    scope: this,
                    params: { type : 'consumption',
                              from : '.meters(all)'},
                    success: function(result, request) {
                        try {
                            var jsonData = Ext.util.JSON.decode(result.responseText);
                            if (jsonData.ok) {
                                var modified = false;

                                Ext.each(jsonData.result.values,
                                         function(val) {
                                            var index = store.find("dsid", new RegExp("^" + val.dsid + "$"));
                                            if (index > -1) {
                                                var rec = store.getAt(index);
                                                rec.set("consumption", val.value);
                                                rec.set("measureDate", val.date);
                                                modified = true;
                                            }
                                         }
                                );
                                if (modified === true) {
                                    store.commitChanges();
                                }

                            } else {
                              Ext.MessageBox.alert('Error', 'Could not get metering data from server');
                            }
                        }
                        catch (err) {
                            Ext.MessageBox.alert('Error', 'Could not get metering data from server');
                        }
                    },
                    failure: function(result, request) {
                        Ext.MessageBox.alert('Error', 'Could not get metering data from server');
                    }
                });
            },
            this
        );
    }
});
Ext.namespace('dSS', 'dSS.data');

dSS.data.DeviceStore = Ext.extend(Ext.data.Store, {
	constructor: function(config) {
		var deviceRecord = Ext.data.Record.create([
			{name:"name"},
			{name:"dsid", mapping: "id"},
			{name:"on"},
			{name:"modulatorDsid"},
			{name:"modulatorName"},
			{name:"zoneId"},
			{name:"isPresent"},
			{name:"firstSeen"},
			{name:"lastDiscovered"}
		]);

		var deviceReader = new Ext.data.JsonReader(
			{
				root:"devices"
			},
			deviceRecord
		);

		Ext.apply(
			this,
			{
				reader: deviceReader,
				sortInfo: {
					field: 'firstSeen',
					direction: 'DESC' // or 'DESC' (case sensitive for local sorting)
				}
			}
		);
		dSS.data.DeviceStore.superclass.constructor.call(this, arguments);
	}
});

Ext.namespace('dSS', 'dSS.data');

dSS.data.LoaderImpl = Ext.extend(Ext.util.Observable, {

  constructor: function() {
    this.modulatorStore = new dSS.data.ModulatorStore({url: '/json/apartment/getCircuits'});
    this.deviceStore = new dSS.data.DeviceStore();
    this.zoneStore = new dSS.data.ZoneStore();
  },

  load: function() {
    this.modulatorStore.load();
    this.loadStructure();
  },

  getModulatorStore: function() {
    return this.modulatorStore;
  },

  getDeviceStore: function() {
    return this.deviceStore;
  },

  getZoneStore: function() {
    return this.zoneStore;
  },


  loadStructure: function() {
    Ext.Ajax.request({
      url: '/json/apartment/getStructure',
      disableCaching: true,
      method: "GET",
      scope: this,
      success: function(result, request) {
        try {
          var jsonData = Ext.util.JSON.decode(result.responseText);
          this.processStructure(jsonData.result);
        }
        catch (err) {
          Ext.MessageBox.alert('AJAX Error', 'Error loading apartment_ structure: "' + err + '"');
        }
      },
      failure: function(result, request) {
        Ext.MessageBox.alert('AJAX Error', 'Could not load "' + request.url + '"');
      }
    });
  },

  processStructure: function(structure) {
    var devices = [], zones = [];
    var zoneStore = this.zoneStore;
    var modulatorStore = this.modulatorStore;
    var deviceStore = this.deviceStore;
    Ext.each(structure.apartment.zones, function(zone) {
      if(zone.id === 0) {
        Ext.each(zone.devices, function(device) {
          device.firstSeen = Date.parseDate(device.firstSeen + "+00:00", "c");
          device.lastDiscovered =  Date.parseDate(device.lastDiscovered + "+00:00", "c");
          devices.push(device);
        });
      }
      zones.push({
        id: zone.id,
        name: zone.name ? zone.name : 'No name',
        primary: ('firstZoneOnDSMeter' in zone) ? true : false
      });
      Ext.each(zone.devices, function(device) {
        for(var i = 0; i < devices.length; i++) {
          if(devices[i].id == device.id) {
            devices[i].zoneId = zone.id;
          }
        }
      });
      zoneStore.loadData({zones: zones});
    });
      Ext.each(devices, function(device) {
        var index = modulatorStore.find("busId", new RegExp("^" + device.circuitID + "$"));
        if(index > -1) {
          var modulator = modulatorStore.getAt(index);
          device.modulatorName = modulator.get("name");
          device.modulatorDsid = modulator.get("dsid")
        }
      });
      deviceStore.loadData({devices: devices});
  }

});

Ext.namespace('dSS', 'dSS.data');

dSS.data.Loader = (function() {
    return new dSS.data.LoaderImpl();
})();


Ext.namespace('dSS');

dSS.ZoneView = Ext.extend(Ext.DataView, {
	initComponent: function() {

		var zoneTemplate = new Ext.XTemplate(
			'<tpl for=".">',
				'<div class="zone-wrap {css}" id="zone-{id}">',
					'<span>',
						'{name}',
						'<tpl if="primary === true">',
							' *',
						'</tpl>',
					'</span>',
				'</div>',
			'</tpl>',
			'<div class="x-clear"></div>'
		);

		Ext.apply(this, {
			store: dSS.data.Loader.getZoneStore(),
			tpl: zoneTemplate,
			singleSelect: true,
			multiSelect: false,
			layout: 'fit',
			style: "overflow: auto",
			overClass:'x-view-over',
			itemSelector:'div.zone-wrap',
			emptyText: 'No Rooms to display'
		});

		dSS.ZoneView.superclass.initComponent.apply(this, arguments);

        dSS.data.Loader.getDeviceStore().on(
            'load',
            this.updateNumDevicesHeader,
            this
        );

		this.on(
			'selectionchange',
			function(dv,nodes) {
				var l = nodes.length;
				var s = l != 1 ? 's' : '';
				this.findParentByType('dsszonepanel').setTitle("Zones (" + l + ' zone' + s + ' selected)');
				this.filterDevices();
                this.updateNumDevicesHeader();
			},
			this
		);

		this.on(
			'render',
			function() {
				var zoneView = this;
				this.dropZone = new Ext.dd.DropZone(zoneView.getEl(), {
					ddGroup          : "zoneDeviceDD",
					getTargetFromEvent: function(e) {
						return e.getTarget(zoneView.itemSelector);
					},
					onNodeEnter : function(target, dd, e, data){
						Ext.fly(target).addClass('my-row-highlight-class');
					},
					onNodeOut : function(target, dd, e, data){
						Ext.fly(target).removeClass('my-row-highlight-class');
					},
					onNodeOver : function(target, dd, e, data){
						return Ext.dd.DropZone.prototype.dropAllowed;
					},
					onNodeDrop : function(target, dd, e, data){
						var record = zoneView.getRecord(target);
						Ext.each(data.selections, function(device) {
							var currentDevice = device;
							Ext.Ajax.request({
								url: '/json/structure/zoneAddDevice',
								disableCaching: true,
								method: "GET",
								scope: zoneView,
								params: {		devid: currentDevice.get('dsid'),
														zone:  record.get('id')
												},
								success: function(result, request) {
									try {
										var jsonData = Ext.util.JSON.decode(result.responseText);
										if(jsonData.ok) {
											currentDevice.set("zoneId", record.data.id);
											currentDevice.store.commitChanges();
                                            this.filterDevices();
                                            this.updateNumDevicesHeader();
										}
									}
									catch (err) {
										Ext.MessageBox.alert('Error', 'Could not move device "' + device.data.dsid + '"');
									}
								},
								failure: function(result, request) {
									Ext.MessageBox.alert('Error', 'Could not move device "' + device.data.dsid + '"');
								}
							});
						});
						return true;
					}
				});
			},
			this
		);

		this.on(
			'contextmenu',
			function(view, index, node, event) {
				event.preventDefault();
				event.stopEvent();
				if(!this.contextMenu) {
					this.contextMenu = new Ext.menu.Menu({});
				} else {
					this.contextMenu.removeAll();
				}

                var menuItem = new Ext.menu.Item({
                    text: 'Rename Room',
                    icon: '/images/page_white_edit.png',
                    handler: this.findParentByType('dsszonepanel').editZone.createDelegate(this.findParentByType('dsszonepanel'), index, true)
                });
				this.contextMenu.add(menuItem)

				var menuItem = new Ext.menu.Item({
					text: 'Delete Room',
					icon: '/images/delete.png',
					handler: this.removeZone.createDelegate(this, index, true)
				});
				this.contextMenu.add(menuItem)

                var menuItem = new Ext.menu.Item({
                    text: 'Blink',
                    icon: '/images/blink.png',
					handler: this.blinkZone.createDelegate(this, index, true)
                });
				this.contextMenu.add(menuItem)

				var xy = event.getXY();
				this.contextMenu.showAt(xy);
			},
			this
		);
	},

    filterDevices: function() {
        var selectedZones = this.getSelectedRecords();
        var store = dSS.data.Loader.getDeviceStore();

        if(selectedZones.length === 0) {
            store.clearFilter();
            return;
        }
        store.filterBy(function(record) {
            for(var i = 0; i < selectedZones.length; i++) {
                if(record.get('zoneId') === selectedZones[i].get('id')) {
                    return true;
                }
            }
            return false;
        });
    },

    updateNumDevicesHeader: function() {
        var count = dSS.data.Loader.getDeviceStore().getRange().length;
        var s = count != 1 ? 's' : '';
        var selectedZones = this.getSelectedRecords();
        var room = '';
        var all = '';
        if(selectedZones.length === 0) {
            all = 'all ';
            room = 's';
        } else {
            room = ' "' + selectedZones[0].get('name') + '"';
        }
        this.findParentByType('dsszonebrowser').devicePanel.setTitle("<span class=\"x-panel-header-text\">Devices in " + all + "room" + room + "</span><span style=\"float: right;\" class=\"x-panel-header-text\">" + count + " Device" + s + "</span>");
    },

	removeZone: function(item, event, index) {
		var record = this.getStore().getAt(index);
		if(record.get('primary') == true) {
			Ext.MessageBox.alert('Error', 'You cannot delete a primary room.');
			return;
		}
		var deviceStore = this.findParentByType('dsszonebrowser').devicePanel.getStore();
		if(deviceStore.query('zone', record.get('id')).getCount() > 0) {
			Ext.MessageBox.alert('Error', 'You cannot delete a non empty room.');
			return;
		}

		Ext.Ajax.request({
			url: '/json/structure/removeZone',
			disableCaching: true,
			method: "GET",
			scope: this,
			params: { zoneID: record.get('id') },
			success: function(result, request) {
				try {
					var jsonData = Ext.util.JSON.decode(result.responseText);
					if(jsonData.ok) {
						this.getStore().removeAt(index);
					} else {
						Ext.MessageBox.alert('Error', 'Could not remove room "' + record.get('name') + '"');
					}
				} catch (err) {
						Ext.MessageBox.alert('Error', 'Could not move device "' + record.get('name') + '"');
				}
			},
			failure: function(result, request) {
				Ext.MessageBox.alert('Error', 'Could not remove room "' + record.get('name') + '"');
			}
		})
	},

	blinkZone: function(item, event, index) {
		var record = this.getStore().getAt(index);

		Ext.Ajax.request({
			url: '/json/zone/blink',
			disableCaching: true,
			method: "GET",
			scope: this,
			params: { id: record.get('id') },
			success: function(result, request) {
				try {
					var jsonData = Ext.util.JSON.decode(result.responseText);
					if(!jsonData.ok) {
						Ext.MessageBox.alert('Error', 'Could not blink room "' + record.get('name') + '"');
					}
				} catch (err) {
					Ext.MessageBox.alert('Error', 'Could not blink room "' + record.get('name') + '"');
				}
			},
			failure: function(result, request) {
				Ext.MessageBox.alert('Error', 'Could not blink room "' + record.get('name') + '"');
			}
		})
	}
});

Ext.reg('dsszoneview', dSS.ZoneView);

Ext.namespace('dSS');

dSS.ZonePanel = Ext.extend(Ext.Panel, {
	initComponent: function() {
		Ext.apply(this, {
			title:'Rooms',
			layout: 'border',
			items: [{ xtype: 'dsszoneview', ref: 'zoneView', region: 'center'}]
		});

		this.tbar = this.buildTopToolbar();

		dSS.ZonePanel.superclass.initComponent.apply(this, arguments);
	},

	buildTopToolbar: function() {
		return [{
				iconCls: 'newZoneAction',
				handler: this.createNewZone,
				scope: this
			},
			{
				iconCls: 'editZoneAction',
				handler: this.editZone,
				scope: this
			},
			'->',
			{
				cls: 'zonePanelMenu',
				menu: [{
                    text: "Reload",
                    handler: this.reload,
                    scope: this
                },
                {
					text: "Rescan",
					handler: this.rescan,
					scope: this
				}]
			}
		];
	},
	rescan: function() {
		this.ownerCt.rescan();
	},
	editZone: function() {
		var zoneStore = this.zoneView.getStore();
		if(this.zoneView.getSelectionCount() !== 1) {
			return;
		} else {
			var record = this.zoneView.getSelectedRecords()[0];
			Ext.Msg.prompt('Rename room', 'New name for room:', function(btn, text){
				if(text !== record.get('name')) {
					Ext.Ajax.request({
						url: '/json/zone/setName',
						disableCaching: true,
						method: "GET",
						params: { id: record.get('id'),
											newName: text},
						success: function(result, request) {
							try {
								var jsonData = Ext.util.JSON.decode(result.responseText);
								if(jsonData.ok) {
									record.set('name', text);
									record.commit();
								} else {
									Ext.MessageBox.alert('Error', 'Could not rename room');
								}
							}
							catch (err) {
								Ext.MessageBox.alert('Error', 'Could not rename room');
							}
						},
						failure: function(result, request) {
							Ext.MessageBox.alert('Error', 'Could not rename room');
						}
					});
				}
			}, this, false, record.get('name'));
		}
	},

	createNewZone: function() {
		var zoneStore = this.zoneView.getStore();
		Ext.Msg.prompt('Create new room', 'Name for new room:', function(btn, text){
			if (btn == 'ok'){
				for(var i = 0; i <= zoneStore.data.length; i++) {
					if(zoneStore.findExact('id', i) === -1) {

						Ext.Ajax.request({
							url: '/json/structure/addZone',
							disableCaching: true,
							method: "GET",
							params: { zoneID: i },
							success: function(result, request) {
								try {
									var jsonData = Ext.util.JSON.decode(result.responseText);
									if(jsonData.ok) {



									Ext.Ajax.request({
										url: '/json/zone/setName',
										disableCaching: true,
										method: "GET",
										params: { id: i,
															newName: text},
										success: function(result, request) {
											try {
												var jsonData = Ext.util.JSON.decode(result.responseText);
												if(jsonData.ok) {
													zoneStore.loadData({zones: [{id: i, name: text}]}, true);
												} else {
													Ext.MessageBox.alert('Error', 'Could not create Zone: ' + json.message);
												}
											}
											catch (err) {
												Ext.MessageBox.alert('Error', 'Could not create Zone: ' + err);
											}
										},
										failure: function(result, request) {
											Ext.MessageBox.alert('Error', 'Could not create Zone');
										}
									});


									}
								}
								catch (err) {
									Ext.MessageBox.alert('Error', 'Could not create Zone');
								}
							},
							failure: function(result, request) {
								Ext.MessageBox.alert('Error', 'Could not create Zone');
							}
						});
						return;
					}
				}
			}
		})
	},
	reload: function() {
		this.ownerCt.loadData();
	}
});

Ext.reg('dsszonepanel', dSS.ZonePanel);

Ext.namespace('dSS', 'dSS.grid');

dSS.grid.DevicePanel = Ext.extend(Ext.grid.GridPanel, {
	initComponent: function() {

		var stateRenderer = function(value, metaData, record, rowIndex, colIndex, store) {
            if(record.get('isPresent') === false) {
                metaData.css = 'isDisabled';
            }
            else {
			    if(record.get('on')) {
				    metaData.css = 'isOn';
			    } else {
				    metaData.css = 'isOff';
			    }
            }
			return String.format('<img class="padding-img" src="{0}"/>',Ext.BLANK_IMAGE_URL);
		};

		var deviceCols = [
			{header: "on", width: 30, resizable: false, sortable: true, dataIndex: 'on', renderer: stateRenderer},
			{header: "dsid",  width: 150, sortable: true, dataIndex: 'dsid', id: 'dsid'},
			{header: "name", width: 150, sortable: true, dataIndex: 'name'},
			{header: "modulator name", width: 100, sortable: true, dataIndex: 'modulatorName' },
			{header: "modulator dsid", width: 150, sortable: true, dataIndex: 'modulatorDsid', id: 'dsid'},
			{header: "zone id", width: 50, sortable: true, dataIndex: 'zoneId'},
			{header: "first seen", width: 150, sortable: true, dataIndex: 'firstSeen', xtype: 'datecolumn', format: 'Y-m-d\\&\\n\\b\\s\\p\\;H:m:i', id: 'firstSeen'},
			{header: "last discovered", width: 150, sortable: true, dataIndex: 'lastDiscovered', xtype: 'datecolumn', format: 'Y-m-d\\&\\n\\b\\s\\p\\;H:m:i', id: 'lastDiscovered'}
		];

		Ext.apply(this, {
			store            : dSS.data.Loader.getDeviceStore(),
			columns          : deviceCols,
			ddGroup          : "zoneDeviceDD",
			enableDragDrop   : true,
			stripeRows       : true,
			forceFit         : true,
			title            : 'Devices',
			viewConfig: {
				autoFill: true,
				getRowClass: function(record, index) {
					var c = record.get('isPresent');
					if (c === false) {
						return 'nonPresentDevice';
					}
					return '';
				}
			}
		});

        this.tbar = this.buildTopToolbar();

		dSS.grid.DevicePanel.superclass.initComponent.apply(this, arguments);

		this.on(
			'rowcontextmenu',
			function(grid, rowIndex, event) {
				event.preventDefault();
				event.stopEvent();
				if(!this.contextMenu) {
					this.contextMenu = new Ext.menu.Menu({});
				} else {
					this.contextMenu.removeAll();
				}

                var sel = this.getSelectionModel();
                if (!sel.isSelected(rowIndex)) {
                    sel.selectRow(rowIndex, false);
                }

                var showRemoveMenuItem = false;
                var showBlinkMenuItem = false;

                var records = sel.getSelections();

                for (var i = 0; i < records.length; i++)
                {
                    var record = records[i];

                    if (record.get('isPresent') === false) {
                        showRemoveMenuItem = true;
                        break;
                    }
                }

				var menuItem = new Ext.menu.Item({
					text: 'Edit Device',
					icon: '/images/page_white_edit.png',
					handler: this.editDevice.createDelegate(this, rowIndex, true)
				});
				this.contextMenu.add(menuItem)

				var menuItem = new Ext.menu.Item({
					text: 'Send Diagnostic Ping',
					icon: '/images/application_osx_terminal.png',
					handler: this.pingDevice.createDelegate(this, rowIndex, true)
				});
				this.contextMenu.add(menuItem)

                if (showRemoveMenuItem === true) {
				    var menuItem = new Ext.menu.Item({
					    text: 'Remove Device',
					    icon: '/images/delete.png',
					    handler: this.removeDevice.createDelegate(this, rowIndex, true)
				    });
                }

                this.contextMenu.add(menuItem);
				var menuItem = new Ext.menu.Item({
				    text: 'Blink',
				    icon: '/images/blink.png',
				    handler: this.blinkDevice.createDelegate(this, rowIndex, true)
				});
                this.contextMenu.add(menuItem);

				var xy = event.getXY();
				this.contextMenu.showAt(xy);
			},
			this
		);

		this.on(
			'cellclick',
			function (grid, rowIndex, columnIndex, event) {
				if(columnIndex === 0) {
					var record = this.getStore().getAt(rowIndex);
                    if (record.get('isPresent') === false) {
                        return;
                    }
					Ext.Ajax.request({
						url: record.get('on') ? '/json/device/turnOff' : '/json/device/turnOn',
						disableCaching: true,
						method: "GET",
						params: { dsid: record.get('dsid') },
						success: function(result, request) {
							try {
								var jsonData = Ext.util.JSON.decode(result.responseText);
								if(jsonData.ok) {
									record.set('on', record.get('on') ? false : true );
									record.commit();
								}
							} catch(err) {
								Ext.MessageBox.alert('Error', err);
							}
						}
					});
				}
			},
			this
		);
	},

    buildTopToolbar: function() {
        return ['->',
        {
            iconCls: 'editDeviceToolbarIcon',
            handler: this.editDevice,
            scope: this
        },{
            iconCls: 'pingDeviceToolbarIcon',
            handler: this.pingDevice,
            scope: this
        },{
            iconCls: 'removeDeviceToolbarIcon',
            handler: this.removeDevice,
            scope: this
        }];
    },

	editDevice: function(item, event, rowIndex) {
        var record;
        var sel = this.getSelectionModel();
        if (!sel.hasSelection())
        {
            Ext.Msg.alert('No device selected!', 'Please select a device from the list!');
            return;
        }

        if (sel.getSelections().length > 1)
        {
            Ext.Msg.alert('Multiple devices selected!', 'It is not possible to edit multiple devices simultaneously, please select only one device!');
            return;
        }

        record = sel.getSelected();

        Ext.Msg.prompt('Edit device', 'Name:', function(btn, text){
			if(text !== record.get('name')) {
				Ext.Ajax.request({
					url: '/json/device/setName',
					disableCaching: true,
					method: "GET",
					params: { dsid: record.get('dsid'),
										newName: text},
					success: function(result, request) {
						try {
							var jsonData = Ext.util.JSON.decode(result.responseText);
							if(jsonData.ok) {
								record.set('name', text);
								record.commit();
							} else {
								Ext.MessageBox.alert('Error', 'Could not rename device');
							}
						}
						catch (err) {
							Ext.MessageBox.alert('Error', 'Could not rename device');
						}
					},
					failure: function(result, request) {
						Ext.MessageBox.alert('Error', 'Could not rename device');
					}
				});
			}
		}, this, false, record.get('name'));
	},

	pingDevice: function(item, event, rowIndex) {
        var record;
        var sel = this.getSelectionModel();
        if (!sel.hasSelection())
        {
            Ext.Msg.alert('No device selected!', 'Please select a device from the list!');
            return;
        }

        if (sel.getSelections().length > 1)
        {
            Ext.Msg.alert('Multiple devices selected!', 'It is not possible to ping multiple devices simultaneously, please select only one device!');
            return;
        }

        record = sel.getSelected();

		Ext.Ajax.request({
			url: '/json/debug/pingDevice',
			disableCaching: true,
			method: "GET",
			params: { dsid: record.get('dsid') },
			success: function(result, request) {
				try {
					var jsonData = Ext.util.JSON.decode(result.responseText);
					if(jsonData.ok) {
						var title = 'Ping ' + record.get('id');
						var message = "HK: " + jsonData.result.qualityHK + "<br/>RK: " + jsonData.result.qualityRK;
						Ext.MessageBox.show({
							title: title,
							msg: message,
							buttons: Ext.Msg.OK,
							minWidth: 300
							});
					} else {
						Ext.MessageBox.alert('Ping Error', jsonData.message);
					}
				}
				catch (err) {
					Ext.MessageBox.alert('Error', 'Could send ping command to dSS: ' + err);
				}
			},
			failure: function(result, request) {
				Ext.MessageBox.alert('Error', 'Could send ping command to dSS');
			}
		});
	},

	removeDevice: function(item, event, rowIndex) {
		var records;
		var store = this.getStore();

		var sel = this.getSelectionModel();
		if (!sel.hasSelection())
		{
			Ext.MessageBox.alert('No device selected!', 'Please select a device from the list!');
			return;
		}
		records = sel.getSelections();

		var skippedPresent = false;
        var notPresent = new Array();

		for (var i = 0; i < records.length; i++)
		{
			var record = records[i];

			if (record.get('isPresent') === true) {
                skippedPresent = true;
				continue;
			}

            notPresent.push(record);
		}

        if (notPresent.length > 0)
            this.deleteDevice(notPresent);
        else {
            if (skippedPresent === true) {
                Ext.MessageBox.alert('Warning', 'Present devices can not be removed!');
            }
        }
	},

	deleteDevice : function(records) {
		var store = this.getStore();
        if (records.length < 1) {
            this.el.unmask();
            return;
        }

        var record = records.shift();

        this.el.mask("Deleting device: " +  record.get("dsid"));
		Ext.Ajax.request({
			url: '/json/structure/removeDevice',
			disableCaching: true,
			method: "GET",
			scope: this,
			params: { devID : record.get('dsid') },
			success: function(result, request) {
				try {
					var jsonData = Ext.util.JSON.decode(result.responseText);
					if(jsonData.ok) {
						store.remove(record);
                        this.deleteDevice(records);
					} else {
                        this.el.unmask();
						Ext.MessageBox.alert('Could not remove', jsonData.message);
					}
				}
				catch (err) {
                    this.el.unmask();
					Ext.MessageBox.alert('Error', 'Could send remove device command to dSS: ' + err);
				}
			},
			failure: function(result, request) {
                this.el.unmask();
				Ext.MessageBox.alert('Error', 'Could send remove device command to dSS');
			}
		});
	},

	blinkDevice: function(item, event, rowIndex) {
		var records;
		var store = this.getStore();

		var sel = this.getSelectionModel();
		if (!sel.hasSelection())
		{
			Ext.MessageBox.alert('No device selected!', 'Please select a device from the list!');
			return;
		}
		records = sel.getSelections();
		for (var i = 0; i < records.length; i++)
		{
		    Ext.Ajax.request({
			    url: '/json/device/blink',
			    disableCaching: true,
			    method: "GET",
			    scope: this,
			    params: { dsid : records[i].get('dsid') },
			    success: function(result, request) {
				    try {
					    var jsonData = Ext.util.JSON.decode(result.responseText);
					    if(!jsonData.ok) {
						    Ext.MessageBox.alert('Could not blink: ', jsonData.message);
					    }
				    }
				    catch (err) {
					    Ext.MessageBox.alert('Error', 'Could send blink device command to dSS: ' + err);
				    }
			    },
			    failure: function(result, request) {
				    Ext.MessageBox.alert('Error', 'Could send blink device command to dSS');
			    }
		    });
		}
	}
});

Ext.reg('dssdevicepanel', dSS.grid.DevicePanel);
Ext.namespace('dSS');

dSS.MessageBusImpl = Ext.extend(Ext.util.Observable, {
	publish: function() {
		dSS.MessageBusImpl.superclass.fireEvent.apply(this, arguments);
	},
	subscribe: function() {
		dSS.MessageBusImpl.superclass.addListener.apply(this, arguments);
	}
});

Ext.namespace('dSS');

dSS.MessageBus = (function() {
    return new dSS.MessageBusImpl();
})();


Ext.namespace('dSS');

dSS.ZoneBrowser = Ext.extend(Ext.Panel, {
	initComponent: function() {

		Ext.apply(this, {
			layout: 'border',
			items: [{
					xtype: 'dsszonepanel',
					ref: 'zonePanel',
					region: 'west',
					width: 225, // give east and west regions a width
					minSize: 175,
					maxSize: 400,
					split: true
				},{
					xtype: 'dssdevicepanel',
					ref: 'devicePanel',
					minSize: 400,
					region: 'center'
				}
			]
		});

		dSS.ZoneBrowser.superclass.initComponent.apply(this, arguments);
		this.on(
			'afterrender',
			this.loadData,
			this
		);

        this.on(
            'activate',
            function() {
                if (this.zonePanel) {
                    this.zonePanel.zoneView.filterDevices();
                }
            },
            this
        );

	},
	allDevices: undefined,
	rescan: function() {
		this.el.mask("rescanning ...");
		Ext.Ajax.request({
			url: '/json/apartment/rescan',
			disableCaching: true,
			method: "GET",
			scope: this,
			success: function(result, request) {
				this.el.unmask();
				this.loadData();
			},
			failure: function(result, request) {
				this.el.unmask();
				Ext.MessageBox.alert('AJAX Error', 'Rescan failed');
			}
		});
	},

	loadData: function() {
		dSS.data.Loader.load();
	}

});

Ext.reg('dsszonebrowser', dSS.ZoneBrowser);
Ext.namespace('dSS', 'dSS.tree');

dSS.tree.PropertyTreeLoader = Ext.extend(Ext.tree.TreeLoader, {
	constructor: function(config) {

		Ext.apply(this, { dataUrl: '/json/property/getChildren', requestMethod: 'GET' });
		dSS.data.DeviceStore.superclass.constructor.call(this, arguments);
	},
	requestData: function(node, callback, scope){
		if(this.fireEvent("beforeload", this, node, callback) !== false){
			this.transId = Ext.Ajax.request({
				method:this.requestMethod,
					url: this.dataUrl||this.url,
					success: this.handleResponse,
					failure: this.handleFailure,
					scope: this,
					argument: {callback: callback, node: node, scope: scope},
					params: this.getParams(node)
				});
		}else{
			this.runCallback(callback, scope || node, []);
		}
	},
	processResponse : function(response, node, callback, scope){
		var json = response.responseText;
		try {
			var o = response.responseData || Ext.decode(json);
			o = o.result;
			node.beginUpdate();
			for(var i = 0, len = o.length; i < len; i++){
				var rawNode = o[i];
				rawNode.text = rawNode.name;
				rawNode.leaf = rawNode.type ===  'none' ? false : true;
				rawNode.path = node.attributes.path === '/' ?
					node.attributes.path + rawNode.name : node.attributes.path + '/' + rawNode.name;
				var n = this.createNode(rawNode);
				if(n){
					node.appendChild(n);
				}
			}
			node.endUpdate();
			this.runCallback(callback, scope || node, [node]);
		}catch(e){
			this.handleFailure(response);
		}
	},
	getParams: function(node) {
		var buf = [];
		buf.push('path', '=', encodeURIComponent(node.attributes.path));
		return buf.join('');
	}
});

Ext.namespace('dSS', 'dSS.system');

dSS.system.PropertyTree = Ext.extend(Ext.Panel, {
	initComponent: function() {

		Ext.apply(this, {
			layout: 'border',
			items: [{
					xtype: 'treepanel',
					ref: 'propertytree',
					region: 'center',
					width: 225, // give east and west regions a width
					minSize: 175,
					maxSize: 400,
					autoScroll: true,
					animate: true,
					enableDD: false,
					containerScroll: true,
					border: false,
					loader: new dSS.tree.PropertyTreeLoader(),
					root: new Ext.tree.AsyncTreeNode({
						expanded: false,
						path: '/',
						type: 'none',
						text: 'dSS',
						leaf: false
					}),
					rootVisible: true,
					listeners: { append: this.handleAppend }
				}
			]
		});

		dSS.system.PropertyTree.superclass.initComponent.apply(this, arguments);
	},
	handleAppend: function(tree, parent, node, index) {
		if (node.attributes.type === 'none') return;
		var url = "/json/property/";
		switch(node.attributes.type) {
			case 'string':
				url += 'getString';
				break;
			case 'integer':
				url += 'getInteger';
				break;
			case 'boolean':
				url += 'getBoolean';
				break;
			default:
				return;
		}
		Ext.Ajax.request({
			url: url,
			params: { path: node.attributes.path },
			method: 'GET',
			success: function(response, opts) {
				var obj = Ext.decode(response.responseText);
				if(obj.ok === true) {
					node.setText(node.text + " : " + obj.result.value);
				}
			},
			failure: function(response, opts) {
				Ext.MessageBox.alert('Error', 'Server side failure with status code ' + response.status);
			}
		});
	}
});

Ext.reg('dsssystempropertytree', dSS.system.PropertyTree);

Ext.namespace('dSS', 'dSS.grid', 'dSS.data');

dSS.grid.ModulatorPanel = Ext.extend(Ext.grid.GridPanel, {
    initComponent: function() {

        var modulatorColumns = [
            {id: 'id', header: "id",  width: 190, sortable: true, dataIndex: "dsid", id: 'dsid' },
            {id: 'name', header: "modulator name", width: 150, sortable: true, dataIndex: 'name'},
            {id: 'consumption', header: "consumption (mW)", width: 130, sortable: true, dataIndex: 'consumption'},
            {id: 'measureDate', header: "measured", width: 150, sortable: true, dataIndex: 'measureDate'}
        ];

        var rowSelect = new Ext.grid.RowSelectionModel({singleSelect : true});

        Ext.apply(this, {
            store            : dSS.data.Loader.getModulatorStore(),
            columns          : modulatorColumns,
            enableDragDrop   : false,
            stripeRows       : true,
            forceFit         : true,
            title            : 'Modulators',
            selModel         : rowSelect,
            viewConfig: {
                autoFill: true,
                getRowClass: function(record, index) {
                    var c = record.get('isPresent');
                    if (c === false) {
                        return 'nonPresentDevice';
                    }
                    return '';
                }
            }
        });

        this.tbar = this.buildTopToolbar();

        dSS.grid.ModulatorPanel.superclass.initComponent.apply(this, arguments);

        this.on(
            'rowcontextmenu',
            function(grid, rowIndex, event) {
                event.preventDefault();
                event.stopEvent();
                if(!this.contextMenu) {
                    this.contextMenu = new Ext.menu.Menu({});
                } else {
                    this.contextMenu.removeAll();
                }

                var menuItem = new Ext.menu.Item({
                    text: 'Edit Modulator',
                    icon: '/images/page_white_edit.png',
                    handler: this.editDevice.createDelegate(this, rowIndex, true)
                });
                this.contextMenu.add(menuItem);

                var menuItem = new Ext.menu.Item({
                    text: 'Rescan Circuit',
                    icon: '/images/page_white_edit.png',
                    handler: this.rescanCircuit.createDelegate(this, rowIndex, true)
                });
                this.contextMenu.add(menuItem);
/*
                if (this.store.getAt(rowIndex).get('isPresent') === false) {
                    var menuItem = new Ext.menu.Item({
                        text: 'Remove Device',
                        icon: '/images/delete.png',
                        handler: this.removeDevice.createDelegate(this, rowIndex, true)
                    });

                    this.contextMenu.add(menuItem);
                }
*/
                var xy = event.getXY();
                this.contextMenu.showAt(xy);
            },
            this
        );

        rowSelect.on(
            'rowselect',
            function(selModel, rowIndex, dataRecord) {
                var store = dSS.data.Loader.getDeviceStore();
                store.filterBy(function(record) {
                    if (record.get('modulatorDsid') === dataRecord.get('dsid')) {
                        return true;
                    }
                    return false;
                });
            },
            this
        );
    },
    buildTopToolbar: function() {
        return ['->',
            {
                iconCls: 'reloadAction',
                handler: this.removeInactiveDevices,
                scope: this
            }]
    },

    editDevice: function(item, event, rowIndex) {
        var record = this.getStore().getAt(rowIndex);
        Ext.Msg.prompt('Edit Modulator', 'Name:', function(btn, text){
            if(text !== record.get('name')) {
                Ext.Ajax.request({
                    url: '/json/circuit/setName',
                    disableCaching: true,
                    method: "GET",
                    params: { id: record.get('dsid'),
                                        newName: text},
                    success: function(result, request) {
                        try {
                            var jsonData = Ext.util.JSON.decode(result.responseText);
                            if(jsonData.ok) {
                                record.set('name', text);
                                record.commit();
                            } else {
                                Ext.MessageBox.alert('Error', 'Could not rename modulator');
                            }
                        }
                        catch (err) {
                            Ext.MessageBox.alert('Error', 'Could not rename modulator');
                        }
                    },
                    failure: function(result, request) {
                        Ext.MessageBox.alert('Error', 'Could not rename modulator');
                    }
                });
            }
        }, this, false, record.get('name'));
    },
    setInitialSelection: function() {
        if (this.getStore().getCount() == 0) {
            this.getStore().clearFilter();
            return;
        }

        var record;
        var index;

        if (this.getSelectionModel().hasSelection()) {
            index = this.getStore().indexOf(this.getSelectionModel().getSelected());
            record = this.getSelectionModel().getSelected();
        } else {
            this.getView().focusRow(0);
            this.getSelectionModel().selectFirstRow();
            record = this.getStore().getAt(0);
            index = 0;
        }

        this.getSelectionModel().fireEvent('rowselect', this.getSelectionModel(), index, record);
    },
    rescanCircuit: function(item, event, rowIndex) {
      var record = this.getStore().getAt(rowIndex);
      Ext.Ajax.request({
        url: '/json/circuit/rescan',
        disableCaching: true,
        method: "GET",
        params: { id: record.get('dsid') },
        success: function(result, request) {
          try {
            var jsonData = Ext.util.JSON.decode(result.responseText);
            if(jsonData.ok) {
            } else {
              Ext.MessageBox.alert('Error', 'Rescan failed');
            }
          }
          catch (err) {
            Ext.MessageBox.alert('Error', 'Rescan failed: ' + err);
          }
        },
        failure: function(result, request) {
          Ext.MessageBox.alert('Error', 'Rescan failed');
        }
    });
  },
  removeInactiveDevices: function() {
    var sel = this.getSelectionModel();
    if (!sel.hasSelection()) {
        Ext.Msg.alert('No modulator selected!', 'Please select a modulator from the list!');
        return;
    }

    record = sel.getSelected();
    var store = dSS.data.Loader.getDeviceStore();

    Ext.Ajax.request({
        url: '/json/structure/removeInactiveDevices',
        disableCaching: true,
        method: "GET",
        scope: this,
        params: { id : record.get('dsid') },
        success: function(result, request) {
            try {
                var jsonData = Ext.util.JSON.decode(result.responseText);
                if(jsonData.ok) {
                    store.each(
                        function(rec) {
                          if (rec.get('modulatorDsid') == record.get('dsid')) {
                            if (rec.get('isPresent') === false) {
                                store.remove(rec);
                            }
                          }
                        },
                        this
                    );
                } else {
                    Ext.MessageBox.alert('Could not remove', jsonData.message);
                }
            }
            catch (err) {
                Ext.MessageBox.alert('Error', 'Could send remove device command to dSS: ' + err);
            }
        },
        failure: function(result, request) {
            Ext.MessageBox.alert('Error', 'Could send remove device command to dSS');
        }
    });


  }
});

Ext.reg('dssmodulatorpanel', dSS.grid.ModulatorPanel);

Ext.namespace('dSS', 'dSS.system', 'dSS.grid');

dSS.system.dSMOverview = Ext.extend(Ext.Panel, {
    initComponent: function() {
        Ext.apply(this, {
            layout: 'border',
            items: [{
                    	xtype: 'dssmodulatorpanel',
                        ref: 'modulatorPanel',
                    	region: 'north',
                        split : true
            },{
                    	xtype: 'dssdevicepanel',
                        region: 'center'
            }]
        });

        dSS.system.dSMOverview.superclass.initComponent.apply(this, arguments);

        this.on(
            'activate',
            function(component) {
               this.modulatorPanel.setInitialSelection();
            },
            this);

        this.on(
            'aftershow',
            function(component) {
                this.modulatorPanel.setHeight(this.getInnerHeight()/2);
            },
            this);
    }
});

Ext.reg('dsmoverview', dSS.system.dSMOverview);

Ext.namespace('dSS', 'dSS.system');

dSS.system.logfiles = Ext.extend(Ext.Panel, {
    initComponent: function() {
        Ext.apply(this, {
            layout: 'table',
            title: 'Logfiles',
            defaults: { border: false },
            layoutConfig: {
                columns: 2,
                tableAttrs: {
                    width: '100%'
                }
            },
            items: [{
                colspan: 2,
                html: '<h2>LOGFILES</h2>'
                    },{
                width: '70%',
                html: '<h2>This is a placeholder</h2><p>Sample text.<br/><br/><br/></p>'
                    },{
                html: '<p class="x-green-button-link"><a href="/js/lib/extjs/resources/images/content_img/ds_logo01.gif">sample link</a></p>'
            },{
                colspan: 2,
                html: '<h2>CONFIGURATION FILES</h2>'
            },{
                html: '<h2>Current system configuration</h2><p>Download current system configuration file.<br/> Use: "Right Click > Save Target As..." to download the file. </p>'
            },{
                html: '<p class="x-green-button-link"><a href="/download/apartment.xml" target=":_blank">apartment.xml</a></p>'
            }]
        });

        dSS.system.logfiles.superclass.initComponent.apply(this, arguments);
    }
});

Ext.reg('systemlogfiles', dSS.system.logfiles);


Ext.namespace('dSS');

dSS.SystemPanel = Ext.extend(Ext.Panel, {
	initComponent: function() {

		var tabRecord = Ext.data.Record.create([
			{ name:"title" }
		]);

		var tabStore = new Ext.data.Store({}, tabRecord);

		var contentPanel = {
			ref: 'contentPanel',
			region: 'center',
			layout: 'card',
			activeItem: 0,
			border: false,
			items: [
				{
					title: 'Property Tree',
					xtype: 'dsssystempropertytree',
					ref: 'systemPropertyTree'
				}, {
					title: 'DSM Overview',
					xtype: 'dsmoverview',
					ref: 'systemdSMOverview'
				}, {
                    title: 'Logfiles',
                    xtype: 'systemlogfiles',
                    ref: 'systemlogfiles'
                }
			]
		};

		Ext.apply(this, {
			layout: 'border',
			items: [ {
					xtype: 'listview',
					title: 'System',
					ref: 'listView',
					store: tabStore,
					singleSelect: true,
					region: 'west',
					hideHeaders: true,
					width: 275,
					columns: [{
						header: 'name',
						dataIndex: 'title'
					}]
				},
				contentPanel
			]
		});

		dSS.SystemPanel.superclass.initComponent.apply(this, arguments);

		this.on(
			'activate',
			function(component) {
				if(this.initializedTabs) {
					this.contentPanel.layout.activeItem.fireEvent("activate", this.contentPanel.layout.activeItem);
					return;
				}
				this.contentPanel.items.each(function(item) {
					var tab = new tabRecord({title: item.title});
					tabStore.add([tab]);
				}, this);
				this.listView.select(0, false, true);
				this.listView.on(
					'activate',
					function() {
					},
					this
				);
				this.listView.on(
					'selectionchange',
					function(listView, selections) {
						this.contentPanel.layout.setActiveItem(selections[0].viewIndex);
						this.contentPanel.layout.activeItem.fireEvent("activate", this.contentPanel.layout.activeItem);
					},
					this
				);
				this.initializedTabs = true;
			},
			this
		);

	}
});

Ext.reg('dsssystempanel', dSS.SystemPanel);
Ext.namespace('dSS');

dSS.GlossaryPanel = Ext.extend(Ext.Panel, {
	initComponent: function() {

		Ext.apply(this, {
			layout: 'border',
			items: [{
					xtype: 'box',
					region: 'center',
					autoEl: {
						tag: 'div',
						html: '<h1>Glossary</h1>'
					}
				}
			]
		});

		dSS.GlossaryPanel.superclass.initComponent.apply(this, arguments);
	}
});

Ext.reg('dssglossarypanel', dSS.GlossaryPanel);

Ext.onReady(function(){
	Ext.get('start').remove();
	var viewport = new Ext.Viewport({
		layout: 'border',
		items: [
			{
				xtype: 'box',
				region: 'north',
				height: 80, // give north and south regions a height
				autoEl: {
					tag: 'div',
					html:'<a href="http://digitalstrom.org/" alt="visit digitalSTROM.org"><img id="logo" src="js/lib/extjs/resources/images/content_img/ds_logo01.gif" alt="dS-Logo"></a>' +
								'<h1><a href="http://digitalstrom.org/" alt="visit digitalSTROM.org">digitalSTROM Setup</a></h1>' +
								'<p id="dssVersion">-</p>'
				}
			}, {
				region: 'center',
				xtype: 'tabpanel',
				activeItem: 0,
				items: [
					{
						title: 'Zones',
						xtype: 'dsszonebrowser',
						ref: 'zoneBrowser'
					},{
						title: 'System Properties',
						xtype: 'dsssystempanel',
						ref: 'systemPanel',
					},{
						title: 'Glossary',
						xtype: 'dssglossarypanel',
						ref: 'glossaryPanel',
					}
				]
			}]
	});

	Ext.Ajax.request({
		url            : "/json/system/version",
		disableCaching : true,
		success        : function(response, options) {
			try{
				var data = Ext.decode(response.responseText);
				if(data.ok === false) {
					throw new Ext.Error('Could not get dSS version!', data.error);
				}
				Ext.get("dssVersion").update(data.message);

			} catch(error) {
			}
		},
		failure        : function(response, options) {}
	});
});
